@@ -8,6 +8,7 @@ import functools
 import lxml.etree
 import html
 import pandas
+import jinjasql
 import xmltodict
 import pyparsing as pp
 import subprocess
@@ -16,6 +17,7 @@ from IPython.core.magic import Magics, magics_class, line_cell_magic
 from IPython.display import display, HTML
 from IPython import get_ipython
 from noteql.schema_queries import queries
+from jinja2.utils import Markup
 
 
 LOCAL_DB_MADE = False
@@ -24,8 +26,6 @@ LOCAL_DB_MADE = False
 def get_engine(dburi):
     if not dburi:
         dburi = os.environ.get("NOTEQL_DBURI")
-        if not dburi:
-            dburi = "postgres://noteql:noteql@db/noteql"
     return sqlalchemy.create_engine(dburi)
 
 
@@ -33,7 +33,12 @@ def generate_rows(result, limit):
     for num, row in enumerate(result):
         if num == limit:
             break
-        yield [json.dumps(item, indent=2) if isinstance(item, dict) else html.escape(str(item)) for item in row]
+        yield [
+            json.dumps(item, indent=2)
+            if isinstance(item, dict)
+            else html.escape(str(item))
+            for item in row
+        ]
 
 
 def double_quote(word):
@@ -73,7 +78,9 @@ def create_local_db():
         service postgresql start
         sudo -u postgres psql -c "CREATE USER root WITH SUPERUSER"
     """
-    process = subprocess.run(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
+    process = subprocess.run(
+        cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE
+    )
     print(process.stderr.decode())
     if process.returncode == 0:
         print("Local DB Made")
@@ -85,8 +92,24 @@ def local_db_session():
     return Session("postgresql+psycopg2://root@/postgres", "public")
 
 
+def safe(value):
+    return Markup(value)
+
+
+def identity(value):
+    return Markup(f""" "{value.replace('"', '""')}" """)
+
+
 class Session:
-    def __init__(self, dburi, schema=None, drop_schema=False, overwrite=False, df_viewer=None, df_viewer_kw=None):
+    def __init__(
+        self,
+        dburi=None,
+        schema=None,
+        drop_schema=False,
+        overwrite=False,
+        df_viewer=None,
+        df_viewer_kw=None,
+    ):
         self.schema = schema
         self.dburi = dburi
         self.df_viewer = df_viewer
@@ -99,12 +122,24 @@ class Session:
 
         self.database_type = self.engine.dialect.name
 
+        param_style = "format"
+        if self.database_type == "sqlite":
+            param_style = "qmark"
+
+        self.jinjarender = jinjasql.JinjaSql(param_style=param_style)
+        self.jinjarender.env.filters["s"] = safe
+        self.jinjarender.env.filters["i"] = identity
+
         self.overwrite = overwrite
         if schema:
             with self.engine.begin() as connection:
                 if drop_schema:
-                    connection.execute("drop schema if exists {} cascade;".format(self.schema))
-                connection.execute("create schema if not exists {};".format(self.schema))
+                    connection.execute(
+                        "drop schema if exists {} cascade;".format(self.schema)
+                    )
+                connection.execute(
+                    "create schema if not exists {};".format(self.schema)
+                )
 
         ipython = get_ipython()
         ipython.register_magics(Noteql)
@@ -128,7 +163,9 @@ class Session:
         return self.get_dataframe(sql, params=[table])
 
     def set(self):
-        print(f'Using db connection {self.dburi} {"schema:" + self.schema if self.schema else ""}')
+        print(
+            f'Using db connection {self.dburi} {"schema:" + self.schema if self.schema else ""}'
+        )
         self.last_set = datetime.datetime.utcnow()
 
     def get_results(self, sql, limit=-1, params=None):
@@ -204,7 +241,7 @@ class Session:
         display(df)
 
     def show_title(self, title, title_size="h3"):
-        display(HTML(f"</br></br><{title_size}>{title}<{title_size}/>"))
+        display(HTML(f"</br><{title_size}>{title}<{title_size}/>"))
 
     def load_json(
         self,
@@ -243,7 +280,11 @@ class Session:
             if self.schema:
                 connection.execute("set local search_path = {};".format(self.schema))
             ## remove quotes when looking at actual table.
-            if table_exists(connection, table_name[1:-1], schema_name[1:-1]) and not overwrite and not append:
+            if (
+                table_exists(connection, table_name[1:-1], schema_name[1:-1])
+                and not overwrite
+                and not append
+            ):
                 print(
                     "WARNING: Table already exists not loading. Set overwrite=True to drop table first or append=True if you want to add rows to existing table"
                 )
@@ -251,7 +292,11 @@ class Session:
             if not append:
                 connection.execute("drop table if exists {}".format(full_name))
 
-            connection.execute('create table if not exists {}(id serial, "{}" jsonb)'.format(full_name, field_name))
+            connection.execute(
+                'create table if not exists {}(id serial, "{}" jsonb)'.format(
+                    full_name, field_name
+                )
+            )
 
             def load(f):
                 if single_cell:
@@ -262,7 +307,11 @@ class Session:
                     print("Total rows loaded 1")
                     return
                 num = -1
-                for num, item in enumerate(ijson.items(f, path_to_list + ("." if path_to_list else "") + "item")):
+                for num, item in enumerate(
+                    ijson.items(
+                        f, path_to_list + ("." if path_to_list else "") + "item"
+                    )
+                ):
                     connection.execute(
                         "insert into {}({}) values (%s)".format(full_name, field_name),
                         json.dumps(item, cls=DecimalEncoder),
@@ -303,7 +352,11 @@ class Session:
         with self.engine.begin() as connection:
             if self.schema:
                 connection.execute("set local search_path = {};".format(self.schema))
-            if table_exists(connection, table_name[1:-1], schema_name[1:-1]) and not overwrite and not append:
+            if (
+                table_exists(connection, table_name[1:-1], schema_name[1:-1])
+                and not overwrite
+                and not append
+            ):
                 print(
                     "WARNING: Table already exists not loading. Set overwrite=True to drop table first or append=True if you want to add rows to existing table"
                 )
@@ -320,7 +373,11 @@ class Session:
                 extra_create += ", {} JSONB".format(json_field)
                 extra_params += ", %s"
 
-            connection.execute('create table if not exists {}("{}" xml {})'.format(full_name, field_name, extra_create))
+            connection.execute(
+                'create table if not exists {}("{}" xml {})'.format(
+                    full_name, field_name, extra_create
+                )
+            )
 
             all_rows = []
             with open(file_name, "rb") as f:
@@ -456,32 +513,63 @@ class DecimalEncoder(json.JSONEncoder):
 @magics_class
 class Noteql(Magics):
     def get_parsers(self):
-        session = (pp.Keyword("session", caseless=True).suppress() + pp.Word(pp.alphanums + "_")("session"))("session")
-        arg = (
-            pp.Word(pp.alphanums)
+        arg_params = (
+            pp.Word(pp.alphanums + "_")
             + pp.Suppress("=")
-            + (pp.Word(pp.alphanums) | pp.QuotedString("'", escQuote="'", unquoteResults=False))
+            + pp.QuotedString("'", escQuote="'")
         )("arg_params")
-        show = pp.Keyword("show", caseless=True)("show")
+
         create = pp.Keyword("create", caseless=True).suppress() + (
-            pp.Word(pp.alphanums + "_") | pp.QuotedString('"', escQuote='"', unquoteResults=False)
+            pp.Word(pp.alphanums + "_")
+            | pp.QuotedString('"', escQuote='"', unquoteResults=False)
         )("create")
-        df = pp.Keyword("df", caseless=True).suppress() + pp.Word(pp.alphanums + "_")("df")
         df_arrows = (pp.Word(pp.alphanums + "_") + pp.Keyword("<<").suppress())("df")
-        sql = pp.Keyword("sql", caseless=True).suppress() + pp.Word(pp.alphanums + "_")("sql")
-        title = pp.Keyword("title", caseless=True).suppress() + pp.QuotedString("'", escQuote="'")("title")
-        params = pp.Keyword("params", caseless=True).suppress() + pp.Word(pp.alphanums + "_")("params")
-        noformat = (pp.Keyword("noformat", caseless=True) | pp.Keyword("nof", caseless=True))("noformat")
+
+        session = (
+            pp.Keyword("session", caseless=True).suppress()
+            + pp.Optional(pp.Keyword("as", caseless=True).suppress())
+            + pp.Word(pp.alphanums + "_")("session")
+        )
+
+        commands = [arg_params, create, df_arrows, session]
+
+        for cmd_string in [
+            "df",
+            "sql",
+            "row",
+            "rows",
+            "col",
+            "cols",
+            "cell",
+            "record",
+            "records",
+            "headings",
+        ]:
+            commands.append(
+                pp.Word(pp.alphanums + "_")(cmd_string)
+                + pp.Suppress("=")
+                + pp.Keyword(cmd_string, caseless=True).suppress()
+            )
+
+        title = pp.Keyword("title", caseless=True).suppress() + pp.QuotedString(
+            "'", escQuote="'"
+        )("title")
+        nojinja = (
+            pp.Keyword("nojinja", caseless=True) | pp.Keyword("noj", caseless=True)
+        )("nojinja")
+        show = pp.Keyword("show", caseless=True)("show")
         rest = pp.Word(pp.printables)("rest")
 
+        commands.extend([title, nojinja, show, rest])
+
         magic_line_parser = pp.ZeroOrMore(
-            pp.Group(arg | session | create | df | df_arrows | params | sql | noformat | title | show | rest),
-            stopOn=pp.LineEnd(),
+            pp.Group(pp.MatchFirst(commands)), stopOn=pp.LineEnd()
         )
 
         cell_parser = pp.OneOrMore(
             pp.SkipTo(
-                (pp.LineStart() + pp.Group(pp.Keyword("%%nql") + magic_line_parser)) | pp.StringEnd(),
+                (pp.LineStart() + pp.Group(pp.Keyword("%%nql") + magic_line_parser))
+                | pp.StringEnd(),
                 include=True,
             )
         )
@@ -510,8 +598,8 @@ class Noteql(Magics):
         session = self.find_session()
 
         actions = {}
-        params = {}
-        format = True
+        arg_params = {}
+        jinja = True
 
         for item in parsed_line:
 
@@ -523,7 +611,9 @@ class Noteql(Magics):
                 param_name = item[0]
                 variable = ns.get(param_name)
                 if not variable:
-                    print(f"Error in %%nql, variable {param_name} does not exist in your notebook")
+                    print(
+                        f"Error in %%nql, variable {param_name} does not exist in your notebook"
+                    )
                     return
                 if not isinstance(variable, Session):
                     print(f"Error in %%nql, variable {param_name} is not a seesion")
@@ -531,76 +621,101 @@ class Noteql(Magics):
                 variable.set()
                 session = variable
 
-            if item.getName() == "params":
-                param_name = item[0]
-                variable = ns.get(param_name)
-                if not variable:
-                    print(f"Error in %%nql, variable {param_name} does not exist in your notebook")
-                    return
-                params = variable
-
             if item.getName() == "arg_params":
                 param_name, param_value = item
-                if param_value[0] == "'":
-                    params[param_name] = param_value[1:-1]
-                else:
-                    variable = ns.get(param_value)
-                    if not variable:
-                        print(f"Error in %%nql, variable {param_name} does not exist in your notebook")
-                        return
-                    if params:
-                        print("Error in %%nql, can not use supplied params with the `PARAMS` argument")
-                        return
-                    params[param_name] = variable
+                arg_params[param_name] = param_value
 
-            if item.getName() == "df":
-                if "df" in actions:
-                    print(f"Error in %%nql, multiple DF statements specified")
-                    return
-                actions["df"] = item[0]
-            if item.getName() == "sql":
-                if "sql" in actions:
-                    print(f"Error in %%nql, multiple SQL statements specified")
-                    return
-                actions["sql"] = item[0]
-            if item.getName() == "create":
-                if "create" in actions:
-                    print(f"Error in %%nql, multiple CREATE statements specified")
-                    return
-                actions["create"] = item[0]
+            for command in [
+                "df",
+                "sql",
+                "create",
+                "row",
+                "rows",
+                "col",
+                "cols",
+                "cell",
+                "record",
+                "records",
+                "headings",
+            ]:
+                if item.getName() == command:
+                    if command in actions:
+                        print(
+                            f"Error in %%nql, multiple {command.upper()} statements specified"
+                        )
+                        return
+                    actions[command] = item[0]
 
             if item.getName() == "show":
                 actions["show"] = True
 
-            if item.getName() == "noformat":
-                format = False
+            if item.getName() == "nojinja":
+                jinja = False
 
             if item.getName() == "title":
                 session.show_title(item[0])
 
-        if format:
-            sql = sql.format(**ns)
+        params = None
+        if jinja:
+            namespace_copy = ns.copy()
+            namespace_copy.update(arg_params)
+            sql, params = session.jinjarender.prepare_query(sql, namespace_copy)
+        if not params:
+            params = None
 
         sql_variable = actions.get("sql")
         if sql_variable:
+            if params:
+                print(f"Error in %%nql, can have params if using SQL command")
+                return
             ns[sql_variable] = sql
 
         if not actions:
             actions["show"] = True
 
-        if not params:
-            params = None
-
         df = None
+
         df_name = actions.get("df")
+        col_name = actions.get("col")
+        cols_name = actions.get("cols")
+        row_name = actions.get("row")
+        rows_name = actions.get("rows")
+        cell_name = actions.get("cell")
+        record_name = actions.get("record")
+        records_name = actions.get("records")
+        headings_name = actions.get("headings")
 
         show = "show" in actions
 
-        if show or df_name:
+        if any([show, df_name, col_name, cols_name, row_name, rows_name, cell_name]):
             df = session.get_dataframe(sql, params=params)
             if df_name:
                 ns[df_name] = df
 
+            if col_name:
+                ns[col_name] = next(iter(df.to_dict("list").values()))
+
+            if cols_name:
+                ns[cols_name] = list(df.to_dict("list").values())
+
+            if row_name:
+                ns[row_name] = df.to_dict("split")["data"][0]
+
+            if rows_name:
+                ns[rows_name] = df.to_dict("split")["data"]
+
+            if cell_name:
+                ns[cell_name] = next(iter(df.to_dict("list").values()))[0]
+
+            if record_name:
+                ns[record_name] = df.to_dict("records")[0]
+
+            if records_name:
+                ns[records_name] = df.to_dict("records")
+
+            if headings_name:
+                ns[headings_name] = list(df.columns)
+
             if show:
                 if session.df_viewer:
                     df = session.df_viewer(df, **session.df_viewer_kw)
@@ -628,8 +743,12 @@ class Noteql(Magics):
                 self.execute_part(parsed_line[1:], sql)
         else:
             ns = self.shell.user_ns
-            sql = line.format(**ns)
+            namespace_copy = ns.copy()
+            params = None
+            sql, params = session.jinjarender.prepare_query(sql, namespace_copy)
+            if not params:
+                params = None
 
             session = self.find_session()
 
-            return session.get_dataframe(sql)
+            return session.get_dataframe(sql, params=params)
